@@ -1,11 +1,11 @@
-// SPDX-License-Identifier: MIT
-/*
- * Xilinx Zynq MPSoC Firmware layer
- *
- *  Copyright (C) 2021 Xilinx, Inc.
- *
- *  Vikram Sreenivasa Batchali <bvikram@xilinx.com>
- */
+/******************************************************************************
+* Copyright (c) 2021 - 2022 Xilinx, Inc.  All rights reserved.
+* Copyright (c) 2022 - 2025, Advanced Micro Devices, Inc. All Rights Reserved.
+* SPDX-License-Identifier: MIT
+*
+* Vikram Sreenivasa Batchali <bvikram@xilinx.com>
+* Sharath Kumar Dasari <sharathk@amd.com>
+******************************************************************************/
 
 #include <fcntl.h>
 #include <mtd/mtd-user.h>
@@ -22,12 +22,16 @@
 #define XST_FAILURE			(0x1)
 
 /* Macros */
-#define SYS_CHECKSUM_OFFSET		(0x3U)
+#define SYS_CHECKSUM_OFFSET			(0x3U)
 #define XBIU_IDEN_STR_OFFSET		(0x24U)
-#define XBIU_IDEN_STR_LEN		(0x4U)
+#define XBIU_IDEN_STR_LEN			(0x4U)
 #define XBIU_QSPI_MFG_INFO_SIZE		(0x100U)
 #define XBIU_IMG_REVISON_OFFSET		(0x70U)
+#define XBIU_IMG_VERSION_OFFSET		(0x9U)
 #define XBIU_IMG_REVISON_SIZE		(0x24U)
+#define XBIU_IMG_VERSION_SIZE		(0x4U)
+#define XBIU_IMG_VERSION_CHECK		(1.03F)
+
 
 /* The below enums denote persistent registers in Qspi Flash */
 struct sys_persistent_state {
@@ -57,21 +61,27 @@ enum sys_boot_img_id {
 static unsigned int calculate_checksum(void);
 static int update_image(char *qspi_mtd_file);
 static int read_image_file(char *input_file);
-static int validate_board_string(void);
-static int update_persistent_registers(char *qspi_mtd_pers_reg_file);
+static int update_nv_registers(char *qspi_mtd_pers_reg_file);
+static int update_persistent_registers(void);
 static void calculate_image_checksum(char *srcaddr, unsigned int len,
 				     unsigned int *calc_crc);
-static int verify_current_running_image(char *qspi_mtd_file);
+static void verify_current_running_image(void);
 static int validate_boot_img_info(void);
-static int print_persistent_state(char *qspi_mtd_file);
+static int read_persistent_register(void);
+static void print_persistent_status(void);
+static char* check_image_update_status(void);
+static int read_mtd_part(char *qspi_mtd_file);
+static char* get_nxt_img_update(void);
 static int print_qspi_mfg_info(void);
 static void print_usage(void);
 static int print_image_rev_info(char *qspi_mtd_file, char *image_name);
 static int clear_multiboot_val(void);
+static int extract_image_version(char *qspi_mtd_file);
 
 /* Variable definitions */
 static char *srcaddr = NULL;
 static unsigned int image_size;
+static float img_ver;
 static struct sys_boot_img_info boot_img_info __attribute__ ((aligned(4U)));
 
 static const unsigned int crc_table[] = {
@@ -163,6 +173,7 @@ int main(int argc, char *argv[])
 {
 	int ret = XST_FAILURE;
 	char qspi_mtd_file[20U] = {0U};
+	char last_boot_img[20U] = {0U};
 	char image_file_name[100U] = {0U};
 	char image_name[8] = {0U};
 	int opt;
@@ -208,10 +219,16 @@ int main(int argc, char *argv[])
 		}
 	}
 
+	ret = read_persistent_register();
+	if (ret != XST_SUCCESS) {
+		return ret;
+	}
+
 	if (help_flag == 1) {
 		print_usage();
 		return XST_SUCCESS;
 	}
+
 	if ((print_flag | verify_flag | update_flag) == 0) {
 		printf("Invalid command format!\n");
 		print_usage();
@@ -219,15 +236,7 @@ int main(int argc, char *argv[])
 	}
 
 	if (print_flag == 1) {
-		ret = print_persistent_state("/dev/mtd2");
-		if (ret != XST_SUCCESS) {
-			printf("Reading persistent registers ");
-			printf("backup\n");
-			ret = print_persistent_state("/dev/mtd3");
-		}
-		if (ret == XST_SUCCESS) {
-			ret = print_qspi_mfg_info();
-		}
+		ret = print_qspi_mfg_info();
 		if (ret != XST_SUCCESS) {
 			return ret;
 		}
@@ -240,30 +249,14 @@ int main(int argc, char *argv[])
 		return XST_SUCCESS;
 	}
 
-
-	/* Validate board string to ensure the app does not run on
-	 * unsupported boards
-	 */
-	ret = validate_board_string();
-	if (ret != XST_SUCCESS)
-		return XST_FAILURE;
-
-	ret = verify_current_running_image("/dev/mtd2");
-	if (ret != XST_SUCCESS) {
-		printf("Reading persistent registers backup\n");
-		ret = verify_current_running_image("/dev/mtd3");
-		if (ret != XST_SUCCESS) {
-			printf("Unable to retrieve persistent registers\n");
-			return XST_FAILURE;
-		}
+	if(update_flag == 1){
+		printf("BootFW image update started\n");
 	}
 
-	printf("Marking last booted image as bootable\n");
-	ret = update_persistent_registers("/dev/mtd2");
-	if (ret < 0)
-		return XST_FAILURE;
+	(void)verify_current_running_image();
 
-	ret = update_persistent_registers("/dev/mtd3");
+	printf("Marking last booted image as bootable\n");
+	ret = update_persistent_registers();
 	if (ret < 0)
 		return XST_FAILURE;
 
@@ -271,7 +264,7 @@ int main(int argc, char *argv[])
 		return ret;
 	}
 
-	printf("Reading Image file\n");
+	printf("Reading BootFW image file\n");
 	ret = read_image_file(image_file_name);
 	if (ret != XST_SUCCESS)
 		goto END;
@@ -281,27 +274,25 @@ int main(int argc, char *argv[])
 	 */
 	if (boot_img_info.persistent_state.last_booted_img ==
 		(char)SYS_BOOT_IMG_A_ID) {
-		printf("Updating Image B\n");
-		strcpy(image_name, "Image B");
+		printf("Updating BootFW image to ImageB bank\n");
+		strcpy(image_name, "ImageB");
 		boot_img_info.persistent_state.img_b_bootable = 0U;
 		strcpy(qspi_mtd_file, "/dev/mtd7");
+		strcpy(last_boot_img, "/dev/mtd5");
 	} else {
-		printf("Updating Image A\n");
-		strcpy(image_name, "Image A");
+		printf("Updating BootFW image to ImageA bank\n");
+		strcpy(image_name, "ImageA");
 		boot_img_info.persistent_state.img_a_bootable = 0U;
 		strcpy(qspi_mtd_file, "/dev/mtd5");
+		strcpy(last_boot_img, "/dev/mtd7");
 	}
 
 	printf("Marking target image as non bootable\n");
-	ret = update_persistent_registers("/dev/mtd2");
+	ret = update_persistent_registers();
 	if (ret < 0)
 		goto END;
 
-	ret = update_persistent_registers("/dev/mtd3");
-	if (ret < 0)
-		goto END;
-
-	printf("Writing Image to %s partition\n",image_name);
+	printf("Writing BootFW image to %s bank\n",image_name);
 	ret = update_image(qspi_mtd_file);
 	if (ret != XST_SUCCESS)
 		goto END;
@@ -315,22 +306,26 @@ int main(int argc, char *argv[])
 		boot_img_info.persistent_state.requested_boot_img =
 			(char)SYS_BOOT_IMG_A_ID;
 	}
-	/* Update persistent register partition */
-	ret = update_persistent_registers("/dev/mtd2");
+	/* Update persistent registers */
+	ret = update_persistent_registers();
 	if (ret < 0)
 		goto END;
 
-	/* Update persistent register backup partition */
-	ret = update_persistent_registers("/dev/mtd3");
-	if (ret < 0)
-		goto END;
-
-	printf("Clearing multiboot register value\n");
-	ret = clear_multiboot_val();
+	ret = extract_image_version(last_boot_img);
 	if (ret != XST_SUCCESS)
 		goto END;
 
-	printf("%s successfully updated to %s partition\n", image_file_name, image_name);
+	if(img_ver < XBIU_IMG_VERSION_CHECK){
+		printf("Clearing multiboot register value\n");
+		ret = clear_multiboot_val();
+		if (ret != XST_SUCCESS)
+			goto END;
+	}
+
+	printf("%s successfully updated to %s bank\n", image_file_name, image_name);
+	printf("Reboot the system to boot the updated BootFW image\n");
+	printf("Mark the BootFW image as bootable using -v option ");
+	printf("on successful boot\n");
 
 END:
 	if (srcaddr)
@@ -363,6 +358,35 @@ static unsigned int calculate_checksum(void)
 	return (0xFFFFFFFFU - checksum);
 }
 
+/*****************************************************************************/
+/**
+ * @brief
+ * This function update both main and backup persistent register
+ * status
+ *
+ * @return	XST_SUCCESS on SUCCESS and XST_FAILURE on failure
+ *
+ *****************************************************************************/
+static int update_persistent_registers(void)
+{
+	int ret = XST_FAILURE;
+
+	/* Update persistent register partition */
+	ret = update_nv_registers("/dev/mtd2");
+	if (ret < 0)
+		return ret;
+
+	/* Update persistent register backup partition */
+	ret = update_nv_registers("/dev/mtd3");
+	if (ret < 0)
+		return ret;
+
+	ret = XST_SUCCESS;
+
+	return ret;
+
+}
+
 /*****************************************************************************/
 /**
  * @brief
@@ -374,7 +398,7 @@ static unsigned int calculate_checksum(void)
  * @return	XST_SUCCESS on SUCCESS and error code on failure
  *
  *****************************************************************************/
-static int update_persistent_registers(char *qspi_mtd_pers_reg_file)
+static int update_nv_registers(char *qspi_mtd_pers_reg_file)
 {
 	int fd_pers_reg, ret = XST_FAILURE;
 	erase_info_t ei = {0U};
@@ -388,7 +412,7 @@ static int update_persistent_registers(char *qspi_mtd_pers_reg_file)
 
 	ret = ioctl(fd_pers_reg, MEMGETINFO, &qspi_mtd_info);
 	if (ret != XST_SUCCESS) {
-		printf("retrieving MTD paartition info failed\n");
+		printf("retrieving MTD partition info failed\n");
 		goto END;
 	}
 
@@ -433,29 +457,8 @@ END:
  * @return	XST_SUCCESS on SUCCESS and error code on failure
  *
  *****************************************************************************/
-static int verify_current_running_image(char *qspi_mtd_file)
+static void verify_current_running_image(void)
 {
-	int fd_pers_reg, ret = XST_FAILURE;
-
-	fd_pers_reg = open(qspi_mtd_file, O_RDONLY);
-	if (fd_pers_reg < 0) {
-		printf("Open Qspi MTD partition failed\n");
-		return ret;
-	}
-
-	ret = read(fd_pers_reg, (char *)&boot_img_info, sizeof(boot_img_info));
-	if (ret != sizeof(boot_img_info)) {
-		printf("Read Qspi MTD partition failed\n");
-		ret = XST_FAILURE;
-		goto END;
-	}
-
-	ret = validate_boot_img_info();
-	if (ret != XST_SUCCESS) {
-		printf("Persistent registers are corrupted\n");
-		goto END;
-	}
-
 	if (boot_img_info.persistent_state.last_booted_img ==
 		(char)SYS_BOOT_IMG_A_ID) {
 		if (boot_img_info.persistent_state.img_a_bootable == 0U)
@@ -465,9 +468,6 @@ static int verify_current_running_image(char *qspi_mtd_file)
 			boot_img_info.persistent_state.img_b_bootable = 1U;
 	}
 
-END:
-	close(fd_pers_reg);
-	return ret;
 }
 
 /*****************************************************************************/
@@ -628,49 +628,6 @@ END:
 	return ret;
 }
 
-/*****************************************************************************/
-/**
- * @brief
- * This function reads the Board revision from eeprom and checks if
- * the application is running on one of the supported boards.
- *
- * @return	XST_SUCCESS if board revision is supported, else XST_FAILURE
- *
- *****************************************************************************/
-static int validate_board_string(void)
-{
-	int ret = XST_FAILURE;
-	FILE *fru;
-	char revision[10U] = {0U};
-	const char *cmd = "ipmi-fru \
-		    --fru-file=/sys/bus/i2c/devices/1-0050/eeprom \
-		    --interpret-oem-data | \
-		    awk -F\": \" \
-		    '/^  *FRU Board Custom*/ { print ($2); exit }'";
-
-	fru = popen(cmd, "r");
-	if (!fru) {
-		printf("Unable to read Board revision from EEprom\n");
-		return ret;
-	}
-	ret = fscanf(fru, "%9s", revision);
-	if (ret < 1) {
-		printf("Unable to read Board revision from EEprom\n");
-		return ret;
-	}
-	if ((strcmp(revision, "A") == 0) || (strcmp(revision, "B") == 0) ||
-	    (strcmp(revision, "Y") == 0) || (strcmp(revision, "Z") == 0) ||
-		(strcmp(revision, "1") == 0) || (strcmp(revision, "2") == 0)) {
-		ret = XST_SUCCESS;
-	} else {
-		printf("Unable to read Board revision from EEprom via ");
-		printf("fru-print utility\n");
-	}
-	pclose(fru);
-
-	return ret;
-}
-
 /*****************************************************************************/
 /**
  * @brief
@@ -725,25 +682,51 @@ static int validate_boot_img_info(void)
 /*****************************************************************************/
 /**
  * @brief
- * This function reads the persistent registers and displays the state of
- * images A and B in a readable format.
+ * This function tries to reads persistent state from backup
+ * persistent partition, if it fails to read from main persistent
+ * partition
+ *
+ * @return	XST_SUCCESS on SUCCESS and XST_FAILURE on failure
+ *
+ *****************************************************************************/
+static int read_persistent_register(void)
+{
+	int ret = XST_FAILURE;
+
+	ret = read_mtd_part("/dev/mtd2");
+	if (ret != XST_SUCCESS) {
+		printf("Reading persistent registers backup\n");
+		ret = read_mtd_part("/dev/mtd3");
+		if (ret != XST_SUCCESS) {
+			printf("Unable to retrieve persistent registers\n");
+		}
+	}
+
+	return ret;
+}
+
+/*****************************************************************************/
+/**
+ * @brief
+ * This function loads persistent register status in to boot_img_info
+ * structure and validates it.
  *
  * @param	qspi_mtd_file denotes the mtd partition to be read
  *
  * @return	XST_SUCCESS on SUCCESS and error code on failure
  *
  *****************************************************************************/
-static int print_persistent_state(char *qspi_mtd_file)
+static int read_mtd_part(char *qspi_mtd_file)
 {
-	int fd_pers_reg, ret = XST_FAILURE;
+	int fd_mtd_part, ret = XST_FAILURE;
 
-	fd_pers_reg = open(qspi_mtd_file, O_RDONLY);
-	if (fd_pers_reg < 0) {
+	fd_mtd_part = open(qspi_mtd_file, O_RDONLY);
+	if (fd_mtd_part < 0) {
 		printf("Open Qspi MTD partition failed\n");
 		return ret;
 	}
 
-	ret = read(fd_pers_reg, (char *)&boot_img_info, sizeof(boot_img_info));
+	ret = read(fd_mtd_part, (char *)&boot_img_info, sizeof(boot_img_info));
 	if (ret != sizeof(boot_img_info)) {
 		printf("Read Qspi MTD partition failed\n");
 		ret = XST_FAILURE;
@@ -756,6 +739,43 @@ static int print_persistent_state(char *qspi_mtd_file)
 		goto END;
 	}
 
+END:
+	close(fd_mtd_part);
+	return ret;
+}
+
+/*****************************************************************************/
+/**
+ * @brief
+ * This function gets next image bank to be updated based on
+ * current persistent status
+ *
+ * @return	string ImageA or ImageB depending on persistent
+ *          register status
+ *
+ *****************************************************************************/
+static char* get_nxt_img_update(void)
+{
+	if (boot_img_info.persistent_state.last_booted_img ==
+		(char)SYS_BOOT_IMG_A_ID)
+		return "ImageB";
+	else
+		return "ImageA";
+}
+
+/*****************************************************************************/
+/**
+ * @brief
+ * This function reads the persistent registers and displays the status of
+ * images A and B in a readable format.
+ *
+ * @param	qspi_mtd_file denotes the mtd partition to be read
+ *
+ * @return	XST_SUCCESS on SUCCESS and error code on failure
+ *
+ *****************************************************************************/
+static void print_persistent_status(void)
+{
 	printf("Image A: ");
 	if (boot_img_info.persistent_state.img_a_bootable == 0U)
 		printf("Non Bootable\n");
@@ -781,10 +801,6 @@ static int print_persistent_state(char *qspi_mtd_file)
 		printf("Image A\n");
 	else
 		printf("Image B\n");
-
-END:
-	close(fd_pers_reg);
-	return ret;
 }
 
 /*****************************************************************************/
@@ -833,6 +849,53 @@ END:
 	return ret;
 }
 
+/*****************************************************************************/
+/**
+ * @brief
+ * This function extracts the version information from
+ * the MTD partition of the provided image
+ *
+ * @param	qspi_mtd_file denotes the mtd partition to be read
+ *
+ * @return	XST_SUCCESS on SUCCESS and error code on failure
+ *
+ *****************************************************************************/
+static int extract_image_version(char *qspi_mtd_file)
+{
+	int fd, ret = XST_FAILURE;
+	char ver_str[XBIU_IMG_REVISON_SIZE + 1U] = {0};
+
+	fd = open(qspi_mtd_file, O_RDONLY);
+	if (fd < 0) {
+		printf("Open Qspi MTD partition failed\n");
+		return ret;
+	}
+
+	ret = lseek(fd, XBIU_IMG_REVISON_OFFSET + XBIU_IMG_VERSION_OFFSET, SEEK_SET);
+	if (ret != XBIU_IMG_REVISON_OFFSET + XBIU_IMG_VERSION_OFFSET) {
+		printf("Seek Qspi MTD partition failed\n");
+		goto END;
+	}
+
+	ret = read(fd, ver_str, XBIU_IMG_VERSION_SIZE);
+	if (ret != XBIU_IMG_VERSION_SIZE) {
+		printf("Read Qspi MTD partition failed\n");
+		ret = XST_FAILURE;
+		goto END;
+	}
+
+	if (ver_str[0U] == 0) {
+		strncpy(ver_str, "Not defined", XBIU_IMG_VERSION_SIZE);
+	}
+
+	img_ver = atof(ver_str);
+	ret = XST_SUCCESS;
+
+END:
+	close(fd);
+	return ret;
+}
+
 /*****************************************************************************/
 /**
  * @brief
@@ -846,6 +909,7 @@ static int print_qspi_mfg_info(void)
 	int fd_pers_reg, ret = XST_FAILURE;
 	char qspi_mfg_info[XBIU_QSPI_MFG_INFO_SIZE + 1U] = {0};
 
+	(void)print_persistent_status();
 	fd_pers_reg = open("/dev/mtd14", O_RDONLY);
 	if (fd_pers_reg < 0) {
 		printf("Open Qspi MTD partition failed\n");
@@ -891,6 +955,33 @@ static int clear_multiboot_val(void)
 	return XST_SUCCESS;
 }
 
+/*****************************************************************************/
+/**
+ * @brief
+ * This function checks if image is updated in previous boot
+ *
+ * @return	string depending on persistent register status
+ *
+ *****************************************************************************/
+static char* check_image_update_status(void)
+{
+	/* Check if image update in previous boot */
+	if (boot_img_info.persistent_state.last_booted_img ==
+		(char)SYS_BOOT_IMG_A_ID) {
+		if(boot_img_info.persistent_state.img_a_bootable == 0U){
+			return "since ImageA\n            bank was updated in previous boot, use this option to mark it\n            as bootable.";
+		}else{
+			return "since no\n            image was updated in previous boot, no need to use this\n            option.";
+		}
+	} else {
+		if(boot_img_info.persistent_state.img_b_bootable == 0U){
+			return "since ImageB\n            bank was updated in previous boot, use this option to mark it\n            as bootable.";
+		}else{
+			return "since no\n            image was updated in previous boot, no need to use this\n            option.";
+		}
+	}
+}
+
 /*****************************************************************************/
 /**
  * @brief
@@ -901,14 +992,12 @@ static int clear_multiboot_val(void)
  *****************************************************************************/
 static void print_usage(void)
 {
-	printf("Usage: sudo image_update -i <path of image file>\n");
-	printf("image_update -i updates qspi image with the image file ");
-	printf("passed as argument.\n");
-	printf("image_update -p prints persistent state registers.\n");
-	printf("image_update -v marks the current running image as ");
-	printf("bootable.\n");
-	printf("image_update -h prints this menu.\n");
-	printf("Can use xmutil bootfw_update instead of image_update in ");
-	printf("any of the above commands.\n");
+	printf("\nUsage: [image_update/xmutil bootfw_update] [option]...\n\n");
+	printf("  -i      updates bootfw image with bootfw bin file passed as argument,\n");
+	printf("            with the current configuration, %s bank would be updated.\n", get_nxt_img_update());
+	printf("  -p      prints persistent status registers.\n");
+	printf("  -v      marks the current running bootfw image as bootable,");
+	printf(" %s\n",check_image_update_status());
+	printf("  -h      prints menu.\n\n");
 }
 
